/**********************************************************************/
/*
     FILE: cwfcmask.c
  PURPOSE: To apply a cloud mask to a CoastWatch file.
   AUTHOR: Peter Hollemans
     DATE: 14/07/1998
  CHANGES: n/a

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfcmask [-b n,n,n ...] cloudmask input output\n\
where: -b n,n,n...      Specifies the cloud mask bits to use for masking,\n\
                        default is all bits.  n ranges from 1 to 8.\n\
       cloudmask	The CoastWatch cloud mask file.\n\
       input            The CoastWatch dataset to mask.\n\
       output           The CoastWatch dataset to write.  The output values\n\
                        will be altered so that cloudy pixels are flagged as\n\
                        invalid data.\n"
#define NARG	3
#define BUFSIZE	512

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfcmask: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int cp (char *a, char *b) {

/* Copies file a to file b.  Returns 0 on success and 1 on failure. */

size_t num_read;
unsigned char buf[BUFSIZE];
FILE *afp, *bfp;

  afp = fopen (a, "rb");			/* open files */
  bfp = fopen (b, "wb");
  if (afp == NULL || bfp == NULL)
    return (1);

                                                /* read/write */
  num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  while (num_read != 0) {
    if (fwrite ((void *) buf, 1, num_read, bfp) < num_read)
      return (1);
    num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  } /* while */

  fclose (afp);					/* finish up */
  fclose (bfp);
  return (0);

} /* cp */

/**********************************************************************/

int main (int argc, char *argv[]) {

unsigned char bits = 0xff;
char *bitstr;
int bit;
char *cfname, *ifname, *ofname;
int cm_id, in_id, out_id, cm_varid, in_varid, out_varid;
int cm_dimid[2], in_dimid[2]; 
char var_name[CW_MAX_NAME];
size_t rows, cols, cm_rows, cm_cols, i, j;
cw_type xtype;
float *fp;
unsigned char *ucp;
size_t start[2], count[2];

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);

  while (argv[0][0] == '-') {
    switch (argv[0][1]) {

      case 'b':
        bits = 0x00;
        bitstr = strtok (argv[1], ",");
        while (bitstr != NULL) {
          bit = atoi (bitstr);
          if (bit < 1 || bit > 8)
            XIT ("invalid bit number", bitstr, -2);
          bits |= (0x01 << (bit-1));
          bitstr = strtok (NULL, ",");
        } /* while */
        argc-=2, argv+=2; 
        break;

      default: 
        err ("invalid option", argv[0]); 
        usage (-2);

    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

  cfname = argv[0];
  ifname = argv[1];
  ofname = argv[2];

						/* open files */
  CHK (cw_open (cfname, CW_NOWRITE, &cm_id), cfname);
  CHK (cw_open (ifname, CW_NOWRITE, &in_id), ifname);

						/* get cloud var info */
  CHK (cw_inq_varid (cm_id, "cloud", &cm_varid), "cloud");
  CHK (cw_inq_var (cm_id, cm_varid, NULL, NULL, NULL, cm_dimid, NULL), 
    "cloud");
  CHK (cw_inq_dim (cm_id, cm_dimid[0], NULL, &cm_rows), "rows");
  CHK (cw_inq_dim (cm_id, cm_dimid[1], NULL, &cm_cols), "columns");
  
						/* get input var info*/
  in_varid = 0;
  CHK (cw_inq_var (in_id, in_varid, var_name, &xtype, NULL, in_dimid, NULL), 
    "variable 0");
  CHK (cw_inq_dim (in_id, in_dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (in_id, in_dimid[1], NULL, &cols), "columns");

  if (xtype != CW_FLOAT)			/* check info */
    XIT ("invalid input variable", var_name, -3);
  if (rows != cm_rows || cols != cm_cols)
    XIT ("files must have same dimensions", "error", -4);

  if (cp (ifname, ofname) != 0)			/* create output file */
    XIT ("file copy failed", "cp", -5);
  CHK (cw_open (ofname, CW_WRITE, &out_id), ofname);
  out_varid = 0;    

						/* allocate memory */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate variable memory", "malloc", -6);
  if ((ucp = (unsigned char *) malloc (cols)) == NULL)
    XIT ("unable to allocate cloud memory", "malloc", -7);

  start[1] = 0;
  count[0] = 1;
  count[1] = cols;

  for (i = 0; i < rows; i++) {

    start[0] = i;				/* read variable/cloud */
    CHK (cw_get_vara_float (in_id, in_varid, start, count, fp), var_name);
    CHK (cw_get_vara_uchar (cm_id, cm_varid, start, count, ucp), "cloud");

    for (j = 0; j < cols; j++) {		/* mask clouds */
      if (bits & ucp[j])
        fp[j] = CW_BADVAL;
    } /* for */

						/* write masked variable */
    CHK (cw_put_vara_float (out_id, out_varid, start, count, fp), var_name);

  } /* for */

  free ((void *) fp);				/* finish up */
  free ((void *) ucp);
  CHK (cw_close (in_id), ifname);
  CHK (cw_close (cm_id), cfname);
  CHK (cw_close (out_id), ofname);

  return (CW_NOERR);

} /* main */

/**********************************************************************/
