/**********************************************************************/
/*
     FILE: cwfcomp.c
  PURPOSE: To create a composite from a series of CoastWatch files.
   AUTHOR: Peter Hollemans
     DATE: 20/10/1998
  CHANGES: 02/11/1998, changed to time order all files, regardless of
           composite type
           08/02/1999, PFH, added -b option
           1999/10/24, PFH, changed output file date to match last file

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfcomp [-c ctype] [-v] [-g ngood] [-b badval] input1 input2\n\
       [input3 [input4 [...]]] output\n\
where: -c ctype         Specifies the composite type.  Selections include\n\
                        mean (default), median, first, last, min, and max.\n\
       -v               Print verbose status messages.\n\
       -g ngood         Minimum number of good pixels required to form an\n\
                        aggregate function, default is 1.\n\
       -b badval        Specifies a floating-point value to treat as invalid\n\
                        data, default is none.\n\
       input1 input2    The CoastWatch datasets to read (at least 2).\n\
       output           The CoastWatch dataset to write.\n"

#define NARG	3
#define BUFSIZE 512

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/* type definitions */

typedef struct {
  char *fname;
  time_t gmt;
} cwftime;

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfcomp: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int cp (char *a, char *b) {

/* Copies file a to file b.  Returns 0 on success and 1 on failure. */

size_t num_read;
unsigned char buf[BUFSIZE];
FILE *afp, *bfp;

  afp = fopen (a, "rb");                        /* open files */
  bfp = fopen (b, "wb");
  if (afp == NULL || bfp == NULL)
    return (1);

                                                /* read/write */
  num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  while (num_read != 0) {
    if (fwrite ((void *) buf, 1, num_read, bfp) < num_read)
      return (1);
    num_read = fread ((void *) buf, 1, BUFSIZE, afp);
  } /* while */

  fclose (afp);                                 /* finish up */
  fclose (bfp);
  return (0);

} /* cp */

/**********************************************************************/

void mean (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a mean value composite row. */

size_t j, k;
double sum;
size_t dgood;

  for (j = 0; j < cols; j++) {

    dgood = 0;					/* calc sum */
    sum = 0;
    for (k = 0; k < n; k++)
      if (ifp[k][j] != CW_BADVAL) {
        dgood++;
        sum += ifp[k][j];
      } /* if */

    if (dgood != 0 && dgood >= ngood)		/* calc mean */
      ofp[j] = sum/dgood;
    else
      ofp[j] = CW_BADVAL;

  } /* for */

} /* mean */

/**********************************************************************/

int compare_float (const void *a, const void *b) {

/* Compares floats for qsort.  Returns < 0 if a < b, > 0 if a > b, or
   0 if a == b. */

float *fa = (float *) a, *fb = (float *) b;

  if (*fa < *fb)
    return (-1);
  else if (*fa > *fb)
    return (1);
  else 
    return (0);

} /* compare_float */

/**********************************************************************/

void median (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a median value composite row. */

size_t j, k;
static float *arr = NULL;
size_t dgood, mid;

  if (arr == NULL)				/* allocate memory */
    arr = (float *) malloc (n*sizeof (float));

  for (j = 0; j < cols; j++) {
 
    dgood = 0;					/* find good vals */
    for (k = 0; k < n; k++)
      if (ifp[k][j] != CW_BADVAL) {
        arr[dgood] = ifp[k][j];
        dgood++;
      } /* if */

    if (dgood != 0 && dgood >= ngood) {		/* calc median */
      qsort ((void *) arr, dgood, sizeof (float), compare_float);
      if (fmod (dgood, 2) == 0)
        ofp[j] = (arr[dgood/2 - 1] + arr[dgood/2]) / 2;
      else
        ofp[j] = arr[(dgood+1)/2 - 1];
    } /* if */
    else
      ofp[j] = CW_BADVAL;

  } /* for */

} /* median */

/**********************************************************************/

void first (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a first value composite row. */

size_t j, k;

  for (j = 0; j < cols; j++) {

    for (k = 0; k < n; k++)			/* find first */
      if (ifp[k][j] != CW_BADVAL) {
        ofp[j] = ifp[k][j];
        break;
      } /* if */

    if (k == n)					/* none found */
     ofp[j] = CW_BADVAL;   

  } /* for */

} /* first */

/**********************************************************************/

void last (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a last value composite row. */

size_t j, k;

  for (j = 0; j < cols; j++) {

    for (k = n-1; k > 0; k--)			/* find last */
      if (ifp[k][j] != CW_BADVAL) {
        ofp[j] = ifp[k][j];
        break;
      } /* if */

    if (k == 0)					/* none found (maybe) */
     ofp[j] = ifp[0][j];

  } /* for */

} /* last */

/**********************************************************************/

void min (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a minimum value composite row. */

size_t j, k;
float min;

  for (j = 0; j < cols; j++) {

    min = CW_BADVAL;				/* find min */
    for (k = 0; k < n; k++)
      if (ifp[k][j] != CW_BADVAL)
        if (min == CW_BADVAL || ifp[k][j] < min)
          min = ifp[k][j];

    ofp[j] = min;				/* assign */

  } /* for */

} /* min */

/**********************************************************************/

void max (float **ifp, float *ofp, size_t cols, size_t n, size_t ngood) {

/* Calculates a maximum value composite row. */

size_t j, k;
float max;

  for (j = 0; j < cols; j++) {

    max = CW_BADVAL;				/* find max */
    for (k = 0; k < n; k++)
      if (ifp[k][j] != CW_BADVAL)
        if (max == CW_BADVAL || ifp[k][j] > max)
          max = ifp[k][j];

    ofp[j] = max;				/* assign */

  } /* for */

} /* max */

/**********************************************************************/

int compare_cwftime (const void *a, const void *b) {

/* Compares the times in CoastWatch files for qsort.  Returns < 0 if a < b,
   > 0 if a > b, or 0 if a == b. */

cwftime *cta = (cwftime *) a, *ctb = (cwftime *) b;

  if (cta->gmt < ctb->gmt)
    return (-1);
  else if (cta->gmt > ctb->gmt)
    return (1);
  else 
    return (0);

} /* compare_cwftime */

/**********************************************************************/

void time_order (char **ifname, size_t n) {

/* Sorts the file names in order of increasing time. */

cwftime *ct;
size_t k;
short year, day, hour_minute, hour, minute, second;
int in_id, varid;
short ydiff;
time_t gmt;

  ct = (cwftime *) malloc (n*sizeof (cwftime));	/* create time table */
  varid = 0;

  for (k = 0; k < n; k++) {
						/* open file */
    CHK (cw_open (ifname[k], CW_NOWRITE, &in_id), ifname[k]);

						/* get date */
    CHK (cw_get_att_short (in_id, varid, "orbit_start_year", &year), 
      "orbit_start_year");
    if (year < 100) {
      if (year >= 70)
        year = 1900 + year;
      else
        year = 2000 + year;
    } /* if */
    CHK (cw_get_att_short (in_id, varid, "orbit_start_day", &day),
      "orbit_start_day");
    CHK (cw_get_att_short (in_id, varid, "orbit_start_hour_minute", 
      &hour_minute), "orbit_start_hour_minute");
    hour = hour_minute/100;
    minute = hour_minute - hour*100;
    CHK (cw_get_att_short (in_id, varid, "orbit_start_second", &second),
      "orbit_start_second");

    ydiff = year - 1970;			/* calc gmt */
    ct[k].gmt = ((long) ydiff*365 + (ydiff+1)/4 + (day-1))*86400.0 + 
      hour*3600.0 + minute*60.0 + second;

    ct[k].fname = ifname[k];			/* copy file name */
    
    CHK (cw_close (in_id), ifname[k])		/* close file */

  } /* for */

						/* sort & copy */
  qsort ((void *) ct, n, sizeof (cwftime), compare_cwftime);
  for (k = 0; k < n; k++)
    ifname[k] = ct[k].fname;

  free ((void *) ct);				/* finish up */

} /* time_order */

/**********************************************************************/

void data_store (FILE **ds, char **ifname, size_t rows, size_t cols, 
  size_t n, float badval, int verbose) {

/* Stores floating point data from the CoastWatch files in a temporary
   binary storage file for later recall.  This avoids having to either
   repeatedly decompress each file or have a large number of files open 
   simultaneously. */

size_t i, j, k;
size_t start[2], count[2];
int in_id, varid;
float *ifp;

  if (((*ds) = tmpfile ()) == NULL)		/* create data storage file */
    XIT ("error creating temporary storage file", "tmpfile", -11);

						/* allocate memory */
  if ((ifp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate input variable memory", "malloc", -12);

  start[1] = 0;					/* read/write data loop */
  count[0] = 1;
  count[1] = cols;
  varid = 0;

  for (k = 0; k < n; k++) {

						/* open file */
    CHK (cw_open (ifname[k], CW_NOWRITE, &in_id), ifname[k]);

    for (i = 0; i < rows; i++) {

      start[0] = i;				/* read */
      CHK (cw_get_vara_float (in_id, varid, start, count, ifp), "variable 0");

      if (badval != CW_BADVAL) {		/* flag bad values */
        for (j = 0; j < cols; j++)
          if (ifp[j] == badval)
            ifp[j] = CW_BADVAL;
      } /* if */

						/* write */
      if (fwrite ((void *) ifp, sizeof (float), cols, *ds) < cols)
        XIT ("error writing to temporary storage file", "fwrite", -13);

    } /* for */

    CHK (cw_close (in_id), ifname[k])		/* close file */

    if (verbose)				/* print status */
      printf ("cwfcomp: finished file [%4d/%4d]\n", k+1, n);

  } /* for */

  free ((void *) ifp);				/* finish up */

} /* data_store */

/**********************************************************************/

void data_read (FILE *ds, size_t rows, size_t cols, size_t i, size_t k,
  float *fp) {

/* Reads data from the temporary binary storage file. */

long offset;

  offset = ((long) k*rows*cols + (long) i*cols) * sizeof (float);
  if (fseek (ds, offset, SEEK_SET) != 0)
    XIT ("error seeking in temporary storage file", "fseek", -21);
  if (fread ((void *) fp, sizeof (float), cols, ds) < cols)
    XIT ("error reading temporary storage file", "fseek", -22);

} /* data_read */

/**********************************************************************/

int main (int argc, char *argv[]) {

char *ctype = "mean";
int verbose = 0;
char **ifname;
char *ofname;
size_t i, k, n;
int in_id, out_id;
int varid;
float **ifp, *ofp;
size_t start[2], count[2];
void (*comp) (float **, float *, size_t, size_t, size_t);
int dimid[2], test_dimid[2];
char var_name[CW_MAX_NAME], test_var_name[CW_MAX_NAME];
cw_type xtype, test_xtype;
size_t rows, cols, test_rows, test_cols;
FILE *ds;
size_t ngood = 1;
float badval = CW_BADVAL;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'c':
        ctype = argv[1];
        argc-=2, argv+=2; 
        break;
      case 'v':
        verbose = 1;
        argc--, argv++;
        break;
      case 'g':
        ngood = atoi (argv[1]);
        argc-=2, argv+=2;
        break;
      case 'b':
        badval = atof (argv[1]);
        argc-=2, argv+=2;
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

  n = argc - 1;					/* get file names */
  ifname = argv;
  ofname = argv[argc-1];  

						/* open first input file */
  CHK (cw_open (ifname[0], CW_NOWRITE, &in_id), ifname[0]);    

  varid = 0;                                    /* get variable info */
  CHK (cw_inq_var (in_id, varid, var_name, &xtype, NULL, dimid, NULL), 
    "variable 0");
  CHK (cw_inq_dim (in_id, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (in_id, dimid[1], NULL, &cols), "columns");

  if (xtype != CW_FLOAT)                        /* check xtype */
    XIT ("invalid input variable", var_name, -3);

  CHK (cw_close (in_id), ifname[0]);		/* close file */

  for (k = 1; k < n; k++) {			/* check other input files */
    CHK (cw_open (ifname[k], CW_NOWRITE, &in_id), ifname[k]);
    CHK (cw_inq_var (in_id, varid, test_var_name, &test_xtype, NULL, 
      test_dimid, NULL), "variable 0");
    CHK (cw_inq_dim (in_id, test_dimid[0], NULL, &test_rows), "rows");
    CHK (cw_inq_dim (in_id, test_dimid[1], NULL, &test_cols), "columns");
    if (strcmp (var_name, test_var_name) != 0)
      XIT ("variables must match", test_var_name, -4);   
    if (rows != test_rows || cols != test_cols)
      XIT ("files must have same dimensions", "error", -5);
    CHK (cw_close (in_id), ifname[k])
  } /* for */

  if (strcmp (ctype, "mean") == 0)		/* check composite type */
    comp = mean;
  else if (strcmp (ctype, "median") == 0)
    comp = median;
  else if (strcmp (ctype, "first") == 0)
    comp = first;
  else if (strcmp (ctype, "last") == 0)
    comp = last;
  else if (strcmp (ctype, "min") == 0)
    comp = min;
  else if (strcmp (ctype, "max") == 0)
    comp = max;
  else
    XIT ("unsupported composite type", ctype, -6);

  time_order (ifname, n);			/* time order files */

  if (cp (ifname[n-1], ofname) != 0)		/* create output file */
    XIT ("output file creation failed", "cp", -7);
  CHK (cw_open (ofname, CW_WRITE, &out_id), ofname);

  if (verbose)					/* init storage file */
    printf ("cwfcomp: reading data into temporary storage\n");
  data_store (&ds, ifname, rows, cols, n, badval, verbose);

  ifp = (float **) malloc (n*sizeof (float *));	/* allocate memory */
  for (k = 0; k < n; k++) {
    if ((ifp[k] = (float *) malloc (sizeof (float)*cols)) == NULL)
      XIT ("unable to allocate input variable memory", "malloc", -8);
  } /* for */
  if ((ofp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate output variable memory", "malloc", -9);

  if (verbose)					/* print status */
    printf ("cwfcomp: creating composite\n");

  start[1] = 0;					/* read/write data loop */
  count[0] = 1;
  count[1] = cols;

  for (i = 0; i < rows; i++) {

    start[0] = i;				/* read */
    for (k = 0; k < n; k++) 
      data_read (ds, rows, cols, i, k, ifp[k]);

    (*comp) (ifp, ofp, cols, n, ngood);		/* composite */

    CHK (cw_put_vara_float (out_id, varid,	/* write */
      start, count, ofp), var_name);

    if (verbose && i % 100 == 0)                /* print status */
      printf ("cwfcomp: finished row  [%4d/%4d]\n", i, rows);

  } /* for */

  CHK (cw_close (out_id), ofname);		/* finish up */
  return (CW_NOERR);

} /* main */

/**********************************************************************/
