/**********************************************************************/
/*
     FILE: cwfnav.c
  PURPOSE: To manually navigate (ie: set shift parameters) for a CoastWatch 
           format file.
   AUTHOR: Peter Hollemans
     DATE: 05/05/1998
  CHANGES: 22/01/1999, PFH, fixed usage message

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwfnav [-h hshift] [-v vshift] [-r] input\n\
where: -h hshift        Changes the default horizontal shift (0).\n\
       -v vshift        Changes the default vertical shift (0).\n\
       -r               Resets the existing shifts before applying new.\n\
       input            The CoastWatch dataset to navigate.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwfnav: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

int main (int argc, char *argv[]) {

short hshift = 0, vshift = 0;
short hshift_old = 0, vshift_old = 0;
int cwid, varid;
int reset = 0;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'h': 
        hshift = atoi (argv[1]);
        argc-=2, argv+=2; 
        break;
      case 'v': 
        vshift = atoi (argv[1]);
        argc-=2, argv+=2; 
        break;
      case 'r':
        reset = 1;
        argc--, argv++; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-3);
  } /* while */

						/* open file */
  CHK (cw_open (argv[0], CW_WRITE, &cwid), argv[0]);

  varid = 0;					/* get attributes */
  if (!reset) {
    CHK (cw_get_att_short (cwid, varid, "horizontal_shift", &hshift_old), 
      "horizontal_shift");
    CHK (cw_get_att_short (cwid, varid, "vertical_shift", &vshift_old), 
      "vertical_shift");
  } /* if */

  hshift += hshift_old;				/* calc new shifts */
  vshift += vshift_old;

						/* put attributes */
  CHK (cw_put_att_short (cwid, varid, "horizontal_shift", CW_SHORT, 1, 
    &hshift), "horizontal_shift");
  CHK (cw_put_att_short (cwid, varid, "vertical_shift", CW_SHORT, 1, 
    &vshift), "vertical_shift");

  CHK (cw_close (cwid), argv[0]);		/* finish up */
  return (CW_NOERR);

} /* main */

/**********************************************************************/
