/**********************************************************************/
/*
     FILE: cwftoarc.c
  PURPOSE: To convert a CoastWatch format file to an ARC/INFO ASCII file
           for input as a grid.
   AUTHOR: Peter Hollemans
     DATE: 09/05/1998
  CHANGES: 26/08/1998, PFH, removed -r option from usage message
           09/10/1998, PFH, corrected lower-left pixel position, cell size
           22/10/1998, PFH, added -b option
           10/12/1998, PFH, fixed fopen "wb" bug for DOS
           15/01/1999, PFH, added polar projection code, added
             projection specifications note

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cwf.h"
#include "cwproj.h"

/* defined constants */

#define UMESSAGE "\
usage: cwftoarc [-p fdig] [-d dfdig] [-b bname] input\n\
where: -p fdig          Changes the default precision (number of\n\
                        significant digits) for floating-point values (6).\n\
       -d ddig          Changes the default precision (number of\n\
                        significant digits) for double precision\n\
                        floating-point values (10).\n\
       -b bname         Changes the output to binary mode, creating\n\
                        <bname>.flt and <bname>.hdr.  Default is ASCII mode\n\
                        to standard output.\n\
       input            The CoastWatch dataset to read.\n"

#define NARG		1
#define PI		3.141592654
#define R_ARC		6370.997e3
#define R_CW		6371.2e3
#define DTOR(a)		((a)*PI/180)
#define MAXLEN		70
#define NAMELEN		80

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

#define LONR(a)	     {\
  if ((a) >= 180)\
    (a) -= 360;\
  else if ((a) < -180)\
    (a) += 360;\
}

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwftoarc: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

void print_val (char *format, float z) {

/* Prints a variable value. */

char cbuf[32];
static int ccount = 0;
int slen;

  sprintf (cbuf, format, z);
  strcat (cbuf, " ");
  slen = strlen (cbuf);
  if (ccount + slen > MAXLEN) {
    printf ("\n");
    ccount = 0;
  } /* if */
  printf (cbuf);
  ccount += slen;

} /* print_val */

/**********************************************************************/

int byte_order (void) {

/* Determines byte order, returning 0 on little endian machines,
   1 on big endian machines, and 2 if the test was inconclusive. */

union {
  long l;
  char c[sizeof (long)];
} test;

  test.l = 1L;
  if (test.c[0] == 1)
    return (0);
  else if (test.c[sizeof(long) - 1] == 1)
    return (1);
  else
   return (2);

} /* byte_order */

/**********************************************************************/

int main (int argc, char *argv[]) {

char format[32];
int cwid, varid, dimid[2];
size_t rows, cols, start[2], count[2], i, j;
cw_type var_xtype;
char var_name[CW_MAX_NAME];
char proj_type[20];
float *fp;
double ll_lat, ll_lon, xllcorner, yllcorner, cellsize;
int fdig = 6, ddig = 10;
float res;
char *bname = NULL;
char hdrname[NAMELEN], fltname[NAMELEN];
FILE *hdrfp, *fltfp, *ofp;
int byte;
short phem;
float plon;
double pdist;
proj_info pinfo;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'p': 
        fdig = atoi (argv[1]); 
        if (fdig < 1)
          XIT ("invalid number of digits", argv[1], -2);
        argc-=2, argv+=2; 
        break;
      case 'd': 
        ddig = atoi (argv[1]); 
        if (ddig < 1)
          XIT ("invalid number of digits", argv[1], -3);
        argc-=2, argv+=2; 
        break;
      case 'b': 
        bname = argv[1];
        argc-=2, argv+=2; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-4);
    } /* switch */
    if (argc < NARG)
      usage (-5);
  } /* while */

						/* open files */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);
  if (bname != NULL) {
    sprintf (hdrname, "%s.%s", bname, "hdr");
    sprintf (fltname, "%s.%s", bname, "flt");
    if ((hdrfp = fopen (hdrname, "w")) == NULL)
      XIT ("can't create header file", hdrname, -6);
    if ((fltfp = fopen (fltname, "wb")) == NULL)
      XIT ("can't create float file", fltname, -6);
  } /* if */

						/* get dimensions */
  CHK (cw_inq_dimid (cwid, "rows", dimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", dimid+1), "columns");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, var_name, &var_xtype, NULL, NULL, NULL),
    "variable 0");

  CHK (cw_init_proj (cwid), argv[0]);		/* initialize projection */
  CHK (cw_get_att_text (cwid, varid, "projection_type", proj_type),
    "projection_type");
  CHK (cw_proj_info (&pinfo), "cw_proj_info");
  res = pinfo.res;
  phem = pinfo.hem;
  plon = pinfo.plon;

  cw_get_ll (0.5, rows+0.5, &ll_lat, &ll_lon);	/* set lower-left corner info */

  switch (pinfo.ptype) {

  case UNMAPPED:				/* ** unmapped ** */
    xllcorner = 0;
    yllcorner = 0;
    cellsize = 1;
    break;

  case POLAR:					/* ** polar ** */
    LONR (ll_lon);
    LONR (plon);
    pdist = 2*R_ARC*tan (PI/4 - phem*DTOR (ll_lat)/2);
    xllcorner = pdist*sin (DTOR (ll_lon - plon));
    yllcorner = -phem*pdist*cos (DTOR (ll_lon - plon));
    cellsize = res*1000.0*(R_ARC/R_CW)*(2/(1 + sin (DTOR (60.0))));
    fprintf (stderr, "cwftoarc: ** NOTE: ArcView Custom Projection **\n");
    fprintf (stderr, "             Projection: Stereographic\n");
    fprintf (stderr, "             Spheroid: Sphere\n");
    fprintf (stderr, "             Central Meridian:   [%8.2f]\n", plon);
    fprintf (stderr, "             Reference Latitude: [%8.2f]\n", 90.0);
    break;

  case MERCATOR:				/* ** mercator ** */
    xllcorner = R_ARC*DTOR(ll_lon);
    yllcorner = R_ARC*log ((1 + sin (DTOR (ll_lat)))/cos (DTOR (ll_lat))); 
    cellsize = res*1000.0*(R_ARC/R_CW);
    fprintf (stderr, "cwftoarc: ** NOTE: ArcView Standard Projection **\n");
    fprintf (stderr, "             Category: Projections of the World\n");
    fprintf (stderr, "             Type: Mercator\n");
    break;

  case LINEAR:					/* ** linear ** */
    xllcorner = ll_lon;
    yllcorner = ll_lat;
    cellsize = res;
    break;

  default:					/* ** unknown ** */
    XIT ("unsupported projection type", proj_type, -6);
    break;

  } /* case */

  if (bname != NULL)				/* write ARC/INFO header */
    ofp = hdrfp;
  else
    ofp = stdout;
  fprintf (ofp, "ncols %u\n", (unsigned int) cols);
  fprintf (ofp, "nrows %u\n", (unsigned int) rows);
  sprintf (format, " %%#.%dg\n", ddig);
  fprintf (ofp, "xllcorner"); fprintf (ofp, format, xllcorner);
  fprintf (ofp, "yllcorner"); fprintf (ofp, format, yllcorner);
  fprintf (ofp, "cellsize"); fprintf (ofp, format, cellsize);
  fprintf (ofp, "nodata_value"); fprintf (ofp, format, CW_BADVAL);

  if (bname != NULL) {			       /* write BYTEORDER */
    byte = byte_order ();
    if (byte == 2)
      XIT ("cannot determine machine byte order", "error", -7);
    fprintf (ofp, "byteorder"); 
    if (byte == 0)
      fprintf (ofp, " LSBFIRST\n");
    else
      fprintf (ofp, " MSBFIRST\n");
  } /* if */    

						/* allocate memory */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -7);


  if (bname == NULL) {				/* write ASCII values */

    if (var_xtype == CW_BYTE)
      sprintf (format, "%%.0f");
    else if (var_xtype == CW_FLOAT)
      sprintf (format, "%%#.%dg", fdig);
    else
      XIT ("unknown", "variable external type", -8);

    start[1] = 0;
    count[0] = 1;
    count[1] = cols;
    for (i = 0; i < rows; i++) {
      start[0] = i;
      CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
      for (j = 0; j < cols; j++)
        print_val (format, fp[j]);
    } /* for */

  } /* if */

  else {					/* write binary values */

    start[1] = 0;
    count[0] = 1;
    count[1] = cols;
    for (i = 0; i < rows; i++) {
      start[0] = i;
      CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
      fwrite ((void *) fp, sizeof (float), cols, fltfp);
    } /* for */

  } /* else */

  free ((void *) fp);				/* finish up */
  if (bname != NULL) {
    fclose (hdrfp);
    fclose (fltfp);
  } /* if */
  CHK (cw_close (cwid), argv[0]);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
