/**********************************************************************/
/*
     FILE: cwftogif.c
  PURPOSE: To plot a CoastWatch format file to a GIF.
   AUTHOR: Peter Hollemans
     DATE: 19/06/1998
  CHANGES: 06/07/1998, PFH, modified for better defaults
           21/08/1998, PFH, added -s and -u options
           31/08/1998, PFH, added dynamic graphics planes
           06/09/1998, PFH, fixed dynamic graphics planes bug
           02/11/1998, PFH, changed default infrared min to 4, added
             "brightness" to non-SST infrared legend
	   28/02/1999, PFH, simplified round function
           29/07/1999, PFH, corrected graphics planes bug, reformatted
             for new C coding style, modified graphics plane layering,
             fixed upper tick mark bug

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

#include "cwf.h"
#include "gd.h"
#include "gdfontg.h"
#include "gdfonts.h"
#include "cmaps.h"

/* defined constants */

#define UMESSAGE "\
usage: cwftogif [-# r,g,b | -# dyn] [-c colors] [-b r,g,b] [-l min] [-h max]\n\
       [-p fdig] [-m cmap] [-t ticint] [-f] [-i] [-s [-u]] [-o output] input\n\
where: -# r,g,b         Specifies the RGB values for graphics plane #\n\
                        (# = 1,2,3,4).\n\
       -# dyn           Specifies that graphics plane # should be dynamic,\n\
                        ie: light on dark, dark on light.\n\
       -c colors        Changes the default number of colors in the\n\
                        palette (245).\n\
       -b r,g,b         Changes the RGB for bad data values (0,0,0).\n\
       -l min           Changes the default minimum data value:\n\
                           infrared = 4\n\
                            visible = 0\n\
                              angle = 0\n\
                          scan time = 0\n\
                              cloud = 0\n\
       -h max           Changes the default maximum data value:\n\
                           infrared = 30\n\
                            visible = 20\n\
                              angle = 90\n\
                          scan time = 24\n\
                              cloud = 255\n\
       -d dec           Changes the default number of decimal places for tick\n\
                        values (1).\n\
       -m cmap          Changes the default color map:\n\
                           infrared = hsl256\n\
                            visible = black-white\n\
                              angle = black-white\n\
                          scan time = black-white\n\
                              cloud = hsl256\n\
                        Selections include:\n\
                          black-white\n\
                          hsl256\n\
                          prism\n\
       -t ticint        Changes the default scale tick interval:\n\
                           infrared = 2\n\
                            visible = 2\n\
                              angle = 5\n\
                          scan time = 2\n\
                              cloud = 16\n\
       -f               Changes the plot to work in Fahrenheit, default is\n\
                        Celsius.\n\
       -i               Inverts the default background and foreground colors\n\
                        (black, white).\n\
       -s               Places a satellite time stamp on the image, default\n\
                        is no stamp.\n\
       -u               Formats the satellite time stamp in UTC, default is\n\
                        the local system time zone.\n\
       -o output        Specifies an output file name, default is write\n\
                        to standard output.\n\
       input            The CoastWatch dataset to read.\n"

#define NARG		1
#define BDR		5
#define BAR		20
#define TIC		5
#define SPC		5
#define MAXCOL		245
#define UTCFORMAT	"%Y/%m/%d %H:%M:%S UTC"
#define LOCFORMAT	"%Y/%m/%d %H:%M:%S %Z"

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

#define CTOF(c) ((c)*1.8 + 32)

#define GRAY(im, c) (0.299*gdImageRed ((im), (c)) + \
  0.587*gdImageGreen ((im), (c)) + \
  0.114*gdImageBlue ((im), (c)))

/* type declarations */

typedef struct {
  int len;
  char *label;
  float frac;
} tic;

/**********************************************************************/

void err (			/*** PRINT ERROR MESSAGE ***/
  const char *m,		/* error message string */
  const char *v			/* variable that message applies to */
) {

  fprintf (stderr, "cwftogif: %s: %s\n", v, m);

} /* err */

/**********************************************************************/

void usage (			/*** PRINT USAGE MESSAGE ***/
  int ecode			/* exit status code */
) {

  fprintf (stderr, UMESSAGE);
  exit (ecode);

} /* usage */

/**********************************************************************/

static int round (		/*** ROUND FLOAT TO NEAREST INTEGER ***/
  float f			/* input float */
) {				/* return: integer result */

  return (f > 0 ? floor (f+0.5) : ceil (f-0.5));

} /* round */

/**********************************************************************/

cmap_data *find_cmap (		/*** FIND COLOR MAP ***/
  char *name			/* name of map to find */
) {				/* return: color map data pointer */

int i;

  /* Search for color map in list */
  /* ---------------------------- */
  for (i = 0; i < NCMAPS; i++)
    if (strcmp (cmaps[i].name, name) == 0)
      return (cmaps+i);

  /* Color map not found */
  /* ------------------- */
  return (NULL);

} /* find_cmap */

/**********************************************************************/

int make_tics (			/*** CREATE TEXT LABELS FOR DATA TICKS ***/
  tic **tics,			/* tick data array (modified) */ 
  float min,			/* minimum data value */
  float max,			/* maximum data value */
  float ticint,			/* tick increment */
  char *format			/* printf-style format for tick label */
) {				/* return: number of ticks created */

int i, ticnum, len;
char cbuf[32];
float span;

  /* Calculate the number of tick marks in the range */
  /* ----------------------------------------------- */
  span = max - min;
  ticnum = round (span/ticint) + 1;
  if (ticnum == 0)
    return (0);

  /* Allocate memory for the tick data */
  /* --------------------------------- */
  *tics = (tic *) malloc (ticnum*sizeof (tic));

  /* Create tick label data */
  /* ---------------------- */
  for (i = 0; i < ticnum; i++) {
    sprintf (cbuf, format, ticint*i + min);
    len = strlen (cbuf);
    (*tics)[i].label = (char *) malloc (len);
    strcpy ((*tics)[i].label, cbuf);
    (*tics)[i].frac = i*ticint / span;
    (*tics)[i].len = len;
  } /* for */

  return (ticnum);

} /* make_tics */

/**********************************************************************/

void draw_borders (		/*** DRAW BORDERS FOR DATA AND COLOR BAR ***/
  gdImagePtr im,		/* gd image pointer */
  int color,			/* gd color map index */
  size_t rows,			/* data rows */
  size_t cols			/* data columns */
) {

int x1, y1, x2, y2;

  /* Draw border around image data */
  /* ----------------------------- */
  x1 = BDR;
  y1 = BDR;
  x2 = x1 + 2*BDR + cols + 1;
  y2 = y1 + 2*BDR + rows + 1;   
  gdImageRectangle (im, x1, y1, x2, y2, color);

  /* Draw border around color bar */
  /* ---------------------------- */
  x1 = x2 + BDR + 1;
  x2 = x1 + 2*BDR + BAR + 1;
  gdImageRectangle (im, x1, y1, x2, y2, color);

} /* draw_borders */

/**********************************************************************/

void draw_tics (		/*** DRAW TICK MARKS AND LABELS ***/
  gdImagePtr im,		/* gd image pointer */
  gdFontPtr font,		/* gd font pointer */
  tic tics[],			/* tick data array */
  int ticnum,			/* number of data ticks */
  int color,			/* gd color map index */
  size_t rows,			/* data rows */
  size_t cols			/* data columns */
) {

int x1, y1, x2, y2, i;

  /* Initialize tick mark width */
  /* -------------------------- */
  x1 = cols + (6*BDR + 4) + BAR;
  x2 = x1 + TIC;

  /* Draw tick marks and labels */
  /* -------------------------- */
  for (i = 0; i < ticnum; i++) {
    y1 = y2 = (gdImageSY (im) - 1) - (2*BDR + 1) - tics[i].frac*(rows - 1);
    gdImageLine (im, x1, y1, x2, y2, color);
    gdImageString (im, font, x2 + SPC, y2 - font->h/2, 
      (unsigned char *) tics[i].label, color);
  } /* for */

} /* draw_tics */ 

/**********************************************************************/

void draw_legend (		/*** DRAW COLOR BAR LEGEND ***/
  gdImagePtr im,		/* gd image pointer */
  gdFontPtr font,		/* gd font pointer */
  int color,			/* gd color map index */
  char *data_id,		/* data id string */
  char *var_name,		/* variable name */ 
  int fahr			/* Fahrenheit flag */
) {

int x1, y1;
char units[80];

  /* Create units string */
  /* ------------------- */
  if (strcmp (data_id, "infrared") == 0) {
    if (strstr (var_name, "sst") == NULL)
      strcpy (units, "brightness ");
    else
      strcpy (units, "");
    if (fahr)
      strcat (units, "temperature ( F)");
    else
      strcat (units, "temperature ( C)");
    units[strlen(units)-3] = 176;
  } /* if */
  else if (strcmp (data_id, "visible") == 0)
    strcpy (units, "albedo (%)");
  else if (strcmp (data_id, "ancillary") == 0) {
    if (strcmp (var_name, "scan_time") == 0)
      strcpy (units, "hours");
    else
      strcpy (units, "degrees");
  } /* else if */
  else if (strcmp (data_id, "cloud") == 0)
    strcpy (units, "byte value");

  /* Draw legend string */
  /* ------------------ */
  x1 = (gdImageSX (im) - 1) - (font->h + SPC);	/* draw string */
  y1 = gdImageSY (im)/2 + strlen (units)*font->w/2;
  gdImageStringUp (im, font, x1, y1, (unsigned char *) units, color);  

} /* draw_legend */

/**********************************************************************/

void draw_image (		/*** DRAW DATA IMAGE ***/
  gdImagePtr im,		/* gd image pointer */
  int cmapa[],			/* gd color map index array */
  int colors,			/* number of colors in cmapa */
  int bad,			/* gd color map index for bad values */
  int gra[],			/* gd color map indices for graphics */
  size_t rows,			/* data rows */
  size_t cols,			/* data columns */
  int cwid,			/* CoastWatch file id */
  int varid,			/* CoastWatch variable id */
  int gvarid,			/* CoastWatch graphics plane id */
  float min,			/* minimum data value */
  float max,			/* maximum data value */
  unsigned char g_rgb,		/* graphics RGB-mode flags */
  unsigned char g_dyn,		/* graphics dynamic-mode flags */
  int fahr,			/* Fahrenheit flag */
  char *var_name,		/* variable name */
  int black,			/* gd color map index for black */
  int white			/* gd color map index for white */
) {

float *fp;
unsigned char *ucp = NULL;
float m, b;
size_t start[2], count[2], i, j, k, layer;
int cval, cindex;
int xoff, yoff;
int nplane[] = {4,1,3,2};

  /* Allocate memory for variable and graphics data */
  /* ---------------------------------------------- */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -1);
  if (g_rgb || g_dyn)
    if ((ucp = (unsigned char *) malloc (cols)) == NULL)
      XIT ("unable to allocate graphics memory", "malloc", -2);

  /* Determine image offsets */
  /* ----------------------- */
  xoff = yoff = (2*BDR) + 1;

  /* Determine scale and offset for data */
  /* ----------------------------------- */
  m = (colors-1)/(max - min);
  b = -m*min;

  /* Initialize data reading parameters */
  /* ---------------------------------- */
  start[1] = 0;
  count[0] = 1;
  count[1] = cols;

  for (i = 0; i < rows; i++) {

    /* Read data and graphics row */
    /* -------------------------- */
    start[0] = i;
    CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
    if (g_rgb || g_dyn)
      CHK (cw_get_vara_uchar (cwid, gvarid, start, count, ucp), "graphics");
  
    for (j = 0; j < cols; j++) {

      /* Flag bad value */
      /* -------------- */
      if (fp[j] == CW_BADVAL)
        cval = bad;

      /* Scale data value */
      /* ---------------- */
      else {
        if (fahr)
          cindex = round (m*CTOF(fp[j]) + b);
        else
          cindex = round (m*fp[j] + b);
        if (cindex < 0)
          cval = bad;
        else if (cindex > colors-1)
          cval = cmapa[colors-1];
        else
          cval = cmapa[cindex];
      } /* else */

      /* Burn in graphics value */
      /* ---------------------- */
      if ((g_rgb || g_dyn) && ((g_rgb & ucp[j]) || (g_dyn & ucp[j]))) {
        for (layer = 0; layer < 4; layer++) {
          k = nplane[layer];
          if CW_GET_G (ucp[j], k) {
            if (CW_GET_G (g_rgb, k))
              cval = gra[k-1];
            else if (CW_GET_G (g_dyn, k))
              cval = (GRAY (im, cval) < 127 ? white : black);
          } /* if */
        } /* for */
      } /* if */

      /* Finally, draw the pixel */
      /* ----------------------- */
      gdImageSetPixel (im, xoff+j, yoff+i, cval);

    } /* for */

  } /* for */

  /* Free memory */
  /* ----------- */
  free ((void *) fp);
  if (ucp == NULL)
    free ((void *) ucp);

} /* draw_image */

/**********************************************************************/

void draw_bar (			/*** DRAW COLOR BAR ***/
  gdImagePtr im,		/* gd image pointer */
  int cmapa[],			/* gd color map index array */
  int colors,			/* number of colors in cmapa */
  size_t rows,			/* data rows */
  size_t cols			/* data columns */
) {

int x1, y1, x2, y2, i, color;

  /* Initialize image offset */
  /* ----------------------- */
  x1 = cols + (5*BDR + 3);
  x2 = x1 + BAR - 1;

  /* Draw color bar line by line */
  /* --------------------------- */
  for (i = 0; i < rows; i++) {
    y1 = y2 = (gdImageSY (im) - 1) - (2*BDR + 1) - i;
    color = round ((float) i/(rows - 1) * (colors - 1));
    gdImageLine (im, x1, y1, x2, y2, cmapa[color]);
  } /* for */

} /* draw_bar */

/**********************************************************************/

void draw_timestamp (		/*** DRAW SATELLITE TIMESTAMP ***/
  gdImagePtr im,		/* gd image pointer */
  gdFontPtr font,		/* gd font pointer */
  int back,			/* gd color map index for background */
  int fore,			/* gd color map index for background */
  int cwid,			/* CoastWatch file id */
  int varid,			/* CoastWatch variable id */
  int utc			/* UTC flag */
) {

short year, day, hour_minute, hour, minute, second;
char sat[20];
char tstring[80], tsstring[80];
short ydiff;
time_t gmt;
int x1, y1, x2, y2;
int height, width;
size_t i, n;
char varname[CW_MAX_NAME];

  /* Get CoastWatch variable name */
  /* ---------------------------- */
  CHK (cw_inq_var (cwid, varid, varname, NULL, NULL, NULL, NULL), 
    "variable 0");
  for (i = 0, n = strlen (varname); i < n; i++) {
    varname[i] = toupper (varname[i]);  
    if (varname[i] == '_') 
      varname[i] = ' ';
  } /* for */ 

  /* Get satellite name */
  /* ------------------ */
  CHK (cw_get_att_text (cwid, varid, "satellite_id", sat), "satellite_id");
  for (i = 0, n = strlen (sat); i < n; i++)
    sat[i] = toupper (sat[i]);  

  /* Get file date and time */
  /* ---------------------- */
  CHK (cw_get_att_short (cwid, varid, "orbit_start_year", &year), 
    "orbit_start_year");
  if (year < 100) {
    if (year >= 70)
      year = 1900 + year;
    else
      year = 2000 + year;
  } /* if */
  CHK (cw_get_att_short (cwid, varid, "orbit_start_day", &day),
    "orbit_start_day");
  CHK (cw_get_att_short (cwid, varid, "orbit_start_hour_minute", &hour_minute),
    "orbit_start_hour_minute");
  hour = hour_minute/100;
  minute = hour_minute - hour*100;
  CHK (cw_get_att_short (cwid, varid, "orbit_start_second", &second),
    "orbit_start_second");

  /* Create timestamp string using date and time */
  /* ------------------------------------------- */
  ydiff = year - 1970;
  gmt = ((long) ydiff*365 + (ydiff+1)/4 + (day-1))*86400.0 + hour*3600.0 + 
    minute*60.0 + second;
  if (utc)
    strftime (tstring, 80, UTCFORMAT, gmtime (&gmt));
  else
    strftime (tstring, 80, LOCFORMAT, localtime (&gmt));
  sprintf (tsstring, "%s %s %s", sat, varname, tstring);

  /* Draw timestamp box */
  /* ------------------ */
  height = font->h;
  width = (strlen (tsstring) + 2)*font->w; 
  x1 = BDR*2 + 1;
  y1 = (gdImageSY (im) - 1) - (BDR*2 + 1) - height;
  x2 = x1 + width;
  y2 = y1 + height;
  gdImageFilledRectangle (im, x1, y1, x2, y2, back);
  gdImageRectangle (im, x1, y1, x2, y2, fore);

  /* Draw timestamp text */
  /* ------------------- */
  gdImageString (im, font, x1 + font->w, y1, 
    (unsigned char *) tsstring, fore);

} /* draw_timestamp */

/**********************************************************************/

int main (int argc, char *argv[]) {

char cbuf[32], format[32];
int gnum;
int grargb[8][3];
int gra[8];
int cmapa[256];
int badrgb[3] = {0,0,0};
int bad;
unsigned char g_rgb = 0;
unsigned char g_dyn = 0;
int colors = MAXCOL;
float min, max;
int setmin = 0, setmax = 0;
int dec = 1;
char *cmap;
int setcmap = 0;
float ticint;
int setticint = 0;
int inverse = 0;
int timestamp = 0;
int utc = 0;
int fahr = 0;
char *ofname = NULL;
int cwid;
FILE *ofp = stdout;
int dimid[2];
size_t rows, cols, i;
int varid, gvarid;
char var_name[CW_MAX_NAME];
char data_id[20];
float defmin, defmax;
char defcmap[20];
cmap_data *cmapd;
float defticint;
int ticnum;
tic *tics;
int maxticlen;
int height, width;
int back, fore, black, white;
int color;

gdImagePtr im;
gdFontPtr font = gdFontGiant;

  /* Parse command line */
  /* ------------------ */
  argc--, argv++;
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {

      case '1': case'2': case'3': case'4':
        cbuf[0] = argv[0][1];
        cbuf[1] = '\0';
        gnum = atoi (cbuf);
        if (strcmp (argv[1], "dyn") == 0)
          g_dyn = CW_PUT_G (g_dyn, gnum);
        else {
          sscanf (argv[1], "%d,%d,%d", &(grargb[gnum-1][0]),
            &(grargb[gnum-1][1]), &(grargb[gnum-1][2]));
          g_rgb = CW_PUT_G (g_rgb, gnum);
        } /* else */
        argc-=2, argv+=2; 
        break;

      case 'c':
        colors = atoi (argv[1]);
        if (colors < 1 || colors > MAXCOL)
          XIT ("invalid number of colors", argv[1], -2)                  
        argc-=2, argv+=2; 
        break;

      case 'b':
        sscanf (argv[1], "%d,%d,%d", badrgb, badrgb+1, badrgb+2);
        argc-=2, argv+=2; 
        break;

      case 'l':
        min = atof (argv[1]);
        setmin = 1;
        argc-=2, argv+=2; 
        break;

      case 'h':
        max = atof (argv[1]);
        setmax = 1;
        argc-=2, argv+=2; 
        break;

      case 'd': 
        dec = atoi (argv[1]); 
        if (dec < 0)
          XIT ("invalid number of decimal places", argv[1], -2);
        argc-=2, argv+=2; 
        break;

      case 'm': 
        cmap = argv[1];
        setcmap = 1;
        argc-=2, argv+=2; 
        break;

      case 't': 
        ticint = atof (argv[1]);
        setticint = 1;
        if (ticint <= 0)
          XIT ("invalid tick interval", argv[1], -2);
        argc-=2, argv+=2; 
        break;

      case 'f':
        fahr = 1;
        argc--, argv++;
        break;       

      case 'i':
        inverse = 1;
        argc--, argv++;
        break;       

      case 's':
        timestamp = 1;
        argc--, argv++;
        break;       

      case 'u':
        utc = 1;
        argc--, argv++;
        break;       

      case 'o':
        ofname = argv[1];
        argc-=2, argv+=2; 
        break;

      default: 
        err ("invalid option", argv[0]); 
        usage (-2);

    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

  /* Open CoastWatch and output GIF files */
  /* ------------------------------------ */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);
  if (ofname != NULL && (ofp = fopen (ofname, "wb")) == NULL)
    XIT ("can't create output file", ofname, -4);

  /* Get variable dimensions */
  /* ----------------------- */
  CHK (cw_inq_dimid (cwid, "rows", dimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", dimid+1), "columns");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");

  /* Get variable information */
  /* ------------------------ */
  varid = 0;
  CHK (cw_inq_var (cwid, varid, var_name, NULL, NULL, NULL, NULL),
    "variable 0");
  if (cw_inq_varid (cwid, "graphics", &gvarid) != CW_NOERR)
    g_rgb = g_dyn = 0;

  /* Determine default min, max, tick increment, and color map */
  /* --------------------------------------------------------- */
  CHK (cw_get_att_text (cwid, varid, "data_id", data_id), "data_id");
  if (strcmp (data_id, "infrared") == 0) {
    defmin = (fahr ? CTOF (4) : 4);
    defmax = (fahr ? CTOF (30) : 30);
    defticint = 2;
    strcpy (defcmap, "hsl256");
  } /* if */
  else if (strcmp (data_id, "visible") == 0) {
    defmin = 0;
    defmax = 20;
    defticint = 2;
    strcpy (defcmap, "black-white");
  } /* else if */
  else if (strcmp (data_id, "ancillary") == 0) {
    if (strcmp (var_name, "scan_time") == 0) {
      defmin = 0;
      defmax = 24;
      defticint = 2;
      strcpy (defcmap, "black-white");
    } /* if */
    else {
      defmin = 0;
      defmax = 90;
      defticint = 5;
      strcpy (defcmap, "black-white");
    } /* else */
  } /* else if */
  else if (strcmp (data_id, "cloud") == 0) {
    defmin = 0;
    defmax = 255;
    defticint = 16;
    strcpy (defcmap, "hsl256");
  } /* else if */
  else
    XIT ("unknown data_id", data_id, -5);

  /* Set min, max, tick increment and color map */
  /* ------------------------------------------ */
  if (!setmin)
    min = defmin;
  if (!setmax)
    max = defmax;
  if (!setcmap)
    cmap = defcmap;
  if (!setticint)
    ticint = defticint;

  /* Check min, max, color map settings */
  /* ---------------------------------- */
  if (min >= max)
    XIT ("invalid data extents", "min >= max", -3);
  cmapd = find_cmap (cmap);
  if (cmapd == NULL)    
    XIT ("invalid color map", cmap, -6);

  /* Create tick labels */
  /* ------------------ */
  sprintf (format, "%%%d.%df", dec+3, dec);
  ticnum = make_tics (&tics, min, max, ticint, format);
  if (ticnum == 0)
    XIT ("ticint too large for min, max", "make_tics", -7);
  maxticlen = tics[0].len;
  for (i = 1; i < ticnum; i++)
    if (tics[i].len > maxticlen)
      maxticlen = tics[i].len;

  /* Call gd library to create overall image */
  /* --------------------------------------- */
  height = rows + (4*BDR + 2);
  width = cols + (6*BDR + 4) + BAR + TIC + (maxticlen*font->w + 2*SPC)
    + (font->h + SPC);
  if ((im = gdImageCreate (width, height)) == NULL)
    XIT ("can't create image", "gdImageCreate", -8);

  /* Allocate gd color map entries */
  /* ----------------------------- */
  if (inverse) {
    back = white = gdImageColorAllocate (im, 255, 255, 255);
    fore = black = gdImageColorAllocate (im, 0, 0, 0);
  } /* if */
  else {
    back = black = gdImageColorAllocate (im, 0, 0, 0);
    fore = white = gdImageColorAllocate (im, 255, 255, 255);
  } /* else */  

  for (i = 0; i < colors; i++) {
    color = round ((float) i/(colors-1) * 255);
    cmapa[i] = gdImageColorAllocate (im, cmapd->rgb[color][0],
      cmapd->rgb[color][1], cmapd->rgb[color][2]);
  } /* for */

  bad = gdImageColorAllocate (im, badrgb[0], badrgb[1], badrgb[2]);

  for (i = 0; i < 8; i++)
    gra[i] = gdImageColorAllocate (im, grargb[i][0], grargb[i][1],
      grargb[i][2]);

  /* Draw various image components */
  /* ----------------------------- */
  draw_borders (im, fore, rows, cols);
  draw_tics (im, font, tics, ticnum, fore, rows, cols);
  draw_legend (im, font, fore, data_id, var_name, fahr);
  draw_image (im, cmapa, colors, bad, gra, rows, cols, cwid, varid, gvarid,
    min, max, g_rgb, g_dyn, fahr, var_name, black, white);
  draw_bar (im, cmapa, colors, rows, cols);
  if (timestamp)
    draw_timestamp (im, gdFontSmall, back, fore, cwid, varid, utc);

  /* Write GIF to disk, close files, and return */
  /* ------------------------------------------ */
  gdImageGif (im, ofp);
  if (ofname != NULL)
    fclose (ofp);
  CHK (cw_close (cwid), argv[0]);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
