/**********************************************************************/
/*
     FILE: cwftonc.c
  PURPOSE: To convert a CoastWatch format file to netCDF (network Common
           Data Form).
   AUTHOR: Peter Hollemans
     DATE: 26/06/1998
  CHANGES: 28/02/1999, PFH, simplified round function

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "cwf.h"
#include "netcdf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwftonc [-i] input output\n\
where: -i               Encode floating-point data as netCDF 16-bit scaled\n\
                        integers, default is 32-bit floating-point.\n\
       input            The CoastWatch dataset to read.\n\
       output           The netCDF dataset to write.\n"

#define NARG	   2
#define SH_BADVAL  -32768
#define FL_BADVAL  -3.4e38
#define SCALE	   0.01

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

#define NSTR(e) nc_strerror ((e))

static int myncerr;
#define NCHK(r,v) { \
  if ((myncerr = (r)) != NC_NOERR) \
    XIT (NSTR (myncerr), (v), myncerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwftonc: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

static int round (float f) {

/* Rounds a float to the nearest integer. */

  return (f > 0 ? floor (f+0.5) : ceil (f-0.5));

} /* round */

/**********************************************************************/

int main (int argc, char *argv[]) {

int cwid, cwdimid[2], cwvarid, cwgvarid, natts;
size_t rows, cols, start[2], count[2], i, j;
cw_type cwatt_xtype, cwvar_xtype;
char att_name[CW_MAX_NAME];
char var_name[CW_MAX_NAME];
int graphics = 0;
char cbuf[32];
char text_val[50];
short short_val;
float float_val;
float *fp;
unsigned char *ucp;
int integer = 0;
short *sp;

int ncid, ncdimid[2], ncvarid, ncgvarid;
nc_type ncvar_xtype;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);

  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case 'i': 
        integer = 1;
        argc--, argv++; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-2);
  } /* while */

						/* open files */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);
  NCHK (nc_create (argv[1], NC_CLOBBER, &ncid), argv[1]);

						/* get dimensions */
  CHK (cw_inq_dimid (cwid, "rows", cwdimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", cwdimid+1), "columns");
  CHK (cw_inq_dim (cwid, cwdimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, cwdimid[1], NULL, &cols), "columns");

						/* define dimensions */
  NCHK (nc_def_dim (ncid, "rows", rows, ncdimid), "rows");
  NCHK (nc_def_dim (ncid, "columns", cols, ncdimid+1), "columns");

  cwvarid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, cwvarid, var_name, &cwvar_xtype, NULL, NULL, &natts), 
    "variable 0");
  if (cw_inq_varid (cwid, "graphics", &cwgvarid) == CW_NOERR)
    graphics = 1;

  if (cwvar_xtype == CW_BYTE)			/* define variables */
    ncvar_xtype = NC_BYTE;
  else if (cwvar_xtype == CW_FLOAT) {
    if (integer) 
      ncvar_xtype = NC_SHORT;
    else        
      ncvar_xtype = NC_FLOAT;
  } /* else if */
  else
    XIT ("unknown", "variable external type", -3);

  NCHK (nc_def_var (ncid, var_name, ncvar_xtype, 2, ncdimid, &ncvarid),
    var_name);
  if (graphics)
    NCHK (nc_def_var (ncid, "graphics", NC_BYTE, 2, ncdimid, &ncgvarid),
      "graphics");

  for (i = 0; i < natts; i++) {			/* define main attributes */
    sprintf (cbuf, "attribute %d", i);
    CHK (cw_inq_attname (cwid, cwvarid, i, att_name), cbuf);
    CHK (cw_inq_att (cwid, cwvarid, att_name, &cwatt_xtype, NULL), att_name);
    switch (cwatt_xtype) {
      case CW_CHAR:
        if (cw_get_att_text (cwid, cwvarid, att_name, text_val) != CW_NOERR)
          strcpy (text_val, "undefined");
        NCHK (nc_put_att_text (ncid, ncvarid, att_name, strlen (text_val),
          text_val), att_name);
        break;
      case CW_SHORT:
        CHK (cw_get_att_short (cwid, cwvarid, att_name, &short_val), att_name);
        NCHK (nc_put_att_short (ncid, ncvarid, att_name, NC_SHORT, 1, 
          &short_val), att_name);
        break;
      case CW_FLOAT:
        CHK (cw_get_att_float (cwid, cwvarid, att_name, &float_val), att_name);
        NCHK (nc_put_att_float (ncid, ncvarid, att_name, NC_FLOAT, 1, 
          &float_val), att_name);
        break;
    } /* switch */
  } /* for */

  if (ncvar_xtype == NC_SHORT) {		/* define other attributes */
    short_val = SH_BADVAL;
    NCHK (nc_put_att_short (ncid, ncvarid, "missing_value", NC_SHORT, 1, 
      &short_val), "missing_value");
    float_val = SCALE;
    NCHK (nc_put_att_float (ncid, ncvarid, "scale_factor", NC_FLOAT, 1, 
      &float_val), "scale_factor");
    float_val = 0;
    NCHK (nc_put_att_float (ncid, ncvarid, "add_offset", NC_FLOAT, 1, 
      &float_val), "add_offset");
  } /* if */       

  else if (ncvar_xtype == NC_FLOAT) {
    float_val = FL_BADVAL;
    NCHK (nc_put_att_float (ncid, ncvarid, "missing_value", NC_FLOAT, 1, 
      &float_val), "missing_value");
  } /* else */

  NCHK (nc_enddef (ncid), argv[1]);		/* end define mode */

						/* allocate memory */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -4);
  if (ncvar_xtype == NC_SHORT) {
    if ((sp = (short *) malloc (sizeof (short)*cols)) == NULL)
      XIT ("unable to allocate data memory", "malloc", -5);
  } /* if */

  start[1] = 0;					/* write data values */
  count[0] = 1;
  count[1] = cols;
  for (i = 0; i < rows; i++) {
    start[0] = i;
    CHK (cw_get_vara_float (cwid, cwvarid, start, count, fp), var_name);

    if (ncvar_xtype == NC_SHORT) {
      for (j = 0; j < cols; j++)
        if (fp[j] == CW_BADVAL)
          sp[j] = SH_BADVAL;
        else
          sp[j] = round (fp[j]/SCALE);
      NCHK (nc_put_vara_short (ncid, ncvarid, start, count, sp), var_name);
    } /* if */

    else if (ncvar_xtype == NC_FLOAT) {
      for (j = 0; j < cols; j++)
        if (fp[j] == CW_BADVAL)
          fp[j] = FL_BADVAL;
      NCHK (nc_put_vara_float (ncid, ncvarid, start, count, fp), var_name);
    } /* else if */
    
    else
      NCHK (nc_put_vara_float (ncid, ncvarid, start, count, fp), var_name);

  } /* for */
  free ((void *) fp);
  if (ncvar_xtype == NC_SHORT)
    free ((void *) sp);

  if (graphics) {

						/* allocate memory */
    if ((ucp = (unsigned char *) malloc (cols)) == NULL)
      XIT ("unable to allocate graphics memory", "malloc", -6);

    for (i = 0; i < rows; i++) {		/* write graphics values */
      start[0] = i;
      CHK (cw_get_vara_uchar (cwid, cwgvarid, start, count, ucp), "graphics");
      NCHK (nc_put_vara_uchar (ncid, ncgvarid, start, count, ucp), "graphics");
    } /* for */
    free ((void *) ucp);

  } /* if */

  CHK (cw_close (cwid), argv[0]);		/* finish up */
  NCHK (nc_close (ncid), argv[1]);
  return (CW_NOERR);

} /* main */    

/**********************************************************************/
