/**********************************************************************/
/*
     FILE: cwftoraw.c
  PURPOSE: To convert a CoastWatch format file to raw bytes.
   AUTHOR: Peter Hollemans
     DATE: 24/11/1997
  CHANGES: 28/01/1998, PFH, replaced size_t with unsigned long for DOS
             TURBO C compiler, added checks for malloc and IMread_data
           05/05/1998, PFH, corrected bug in round_float routine (it was
             rounding negative numbers incorrectly)
           07/05/1998, PFH, rewrote to use new libcwf
           28/02/1999, PFH, simplified round function

  CoastWatch Format (CWF) Software Library and Utilities
  Copyright 1998, USDOC/NOAA/NESDIS CoastWatch West Coast Regional Node

*/
/**********************************************************************/

/* include files */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cwf.h"

/* defined constants */

#define UMESSAGE "\
usage: cwftoraw [-# gbyte] [-m maxbyte] [-b badbyte] [-o output] [-l min]\n\
       [-h max] input \n\
where: -# gbyte         Specifies the byte value for graphics plane #\n\
                        (# = 1,2,3,4).\n\
       -m maxbyte       Changes the default maximum byte value (255).\n\
       -b badbyte       Changes the default bad data byte value (0).\n\
       -o output        Specifies an output file name, default is write\n\
                        to standard output.\n\
       -l min           Changes the default minimum data value (0.0).\n\
       -h max           Changes the default maximum data value (25.5).\n\
       input            The CoastWatch dataset to read.\n"

#define NARG	1

/* defined macros */

#define XIT(m,v,e) { \
  err ((m), (v)); \
  exit ((e)); \
} 

#define STR(e) cw_strerror ((e))

static int cwerr;
#define CHK(r,v) { \
  if ((cwerr = (r)) != CW_NOERR) \
    XIT (STR (cwerr), (v), cwerr); \
} 

/**********************************************************************/

void err (const char *m, const char *v) {
  fprintf (stderr, "cwftoraw: %s: %s\n", v, m);
} /* err */

/**********************************************************************/

void usage (int ecode) {
  fprintf (stderr, UMESSAGE);
  exit (ecode);
} /* usage */

/**********************************************************************/

static int round (float f) {

/* Rounds a float to the nearest integer. */

  return (f > 0 ? floor (f+0.5) : ceil (f-0.5));

} /* round */

/**********************************************************************/

int main (int argc, char *argv[]) {

unsigned char gbyte[8];
unsigned char g_on = 0;
unsigned gnum;
char cbuf[32];
unsigned char maxbyte = 255;
unsigned char badbyte = 0;
int cwid, varid, gvarid, dimid[2];
size_t rows, cols, start[2], count[2], i, j, k;
char var_name[CW_MAX_NAME];
float m, b;
float min = 0.0, max = 25.5;
unsigned char outbyte;
int ioutbyte;
FILE *ofp = stdout;
char *ofname = NULL;
float *fp;
unsigned char *ucp = NULL;

  argc--, argv++;				/* parse command line */
  if (argc < NARG)
    usage (-1);
  while (argv[0][0] == '-') {
    switch (argv[0][1]) {
      case '1': case'2': case'3': case'4':
        cbuf[0] = argv[0][1];
        cbuf[1] = '\0';
        gnum = (unsigned char) atoi (cbuf);
        gbyte[gnum] = (unsigned char) atoi (argv[1]);
        g_on = CW_PUT_G (g_on, gnum);
        argc-=2, argv+=2; 
        break;
      case 'm':
        maxbyte = (unsigned char) atoi (argv[1]);
        argc-=2, argv+=2; 
        break;
      case 'b':
        badbyte = (unsigned char) atoi (argv[1]);
        argc-=2, argv+=2; 
        break;
      case 'o':
        ofname = argv[1];
        argc-=2, argv+=2; 
        break;
      case 'l':
        min = atof (argv[1]);
        argc-=2, argv+=2; 
        break;
      case 'h':
        max = atof (argv[1]);
        argc-=2, argv+=2; 
        break;
      default: 
        err ("invalid option", argv[0]); 
        usage (-2);
    } /* switch */
    if (argc < NARG)
      usage (-3);
  } /* while */

  if (min == max)				/* check min/max */
    XIT ("values cannot be equal", "min == max", -7);

						/* open files */
  CHK (cw_open (argv[0], CW_NOWRITE, &cwid), argv[0]);
  if (ofname != NULL && (ofp = fopen (ofname, "wb")) == NULL)
    XIT ("can't create output file", ofname, -4);

						/* get dimensions */
  CHK (cw_inq_dimid (cwid, "rows", dimid), "rows");
  CHK (cw_inq_dimid (cwid, "columns", dimid+1), "columns");
  CHK (cw_inq_dim (cwid, dimid[0], NULL, &rows), "rows");
  CHK (cw_inq_dim (cwid, dimid[1], NULL, &cols), "columns");

  varid = 0;					/* get variable info */
  CHK (cw_inq_var (cwid, varid, var_name, NULL, NULL, NULL, NULL),
    "variable 0");
  if (cw_inq_varid (cwid, "graphics", &gvarid) != CW_NOERR)
    g_on = 0;

						/* allocate memory */
  if ((fp = (float *) malloc (sizeof (float)*cols)) == NULL)
    XIT ("unable to allocate data memory", "malloc", -5);
  if (g_on)
    if ((ucp = (unsigned char *) malloc (cols)) == NULL)
      XIT ("unable to allocate graphics memory", "malloc", -6);

  m = (float) maxbyte/(max - min);	        /* determine scale, offset */
  b = -m*min;

  start[1] = 0;					/* read/write data loop */
  count[0] = 1;
  count[1] = cols;

  for (i = 0; i < rows; i++) {

    start[0] = i;				/* read data/graphics */
    CHK (cw_get_vara_float (cwid, varid, start, count, fp), var_name);
    if (g_on)
      CHK (cw_get_vara_uchar (cwid, gvarid, start, count, ucp), "graphics");
  
    for (j = 0; j < cols; j++) {

      if ((g_on) && (g_on & ucp[j])) {		/* burn in graphics */
        for (k = 4; k >= 1; k--) {
          if (CW_GET_G (g_on, k) && CW_GET_G (ucp[j], k))
            outbyte = gbyte[k];
        } /* for */
      } /* if */

      else if (fp[j] == CW_BADVAL)		/* detect badval */
        outbyte = badbyte;

      else {					/* scale data value */
        ioutbyte = round (m*fp[j] + b);
        if (ioutbyte < 0)
          outbyte = 0;
        else if (ioutbyte > (int) maxbyte)
          outbyte = maxbyte;
        else
          outbyte = (unsigned char) ioutbyte;
      } /* else */

						/* write raw byte */
      if (fwrite ((void *) &outbyte, 1, 1, ofp) < 1)
        XIT ("error writing output", "fwrite", -8);

    } /* for */

  } /* for */

  free ((void *) fp);				/* finish up */
  if (ucp == NULL)
    free ((void *) ucp);
  if (ofname != NULL)
    fclose (ofp);
  CHK (cw_close (cwid), argv[0]);
  return (CW_NOERR);

} /* main */

/**********************************************************************/
